/* be_x86.c
 *
 * backend for x86 code generation (uses nasm as assembler)
 * this is the old 8 bit version
 *
 * Copyright (C) 2004 Manuel Schiller
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <assert.h>
#include "bf.h"

#define MAXNEST	1024	/* loops can not be nested deeper than this */

int gen_x86(FILE *out, bfinstr_t *pgm)
{
	long label = 0;
	long lstk[MAXNEST];
	long ltop = 0;

	fprintf(out, "section .text\nglobal _start\n_start:" \
		     /* register setup */
		     "\txor eax, eax\n\txor edx,edx\n\n\tinc edx\n" \
		     /* allocate memory on the stack and clear it,
		      * allocate a little more because of buggy brainfuck
		      * programs which underflow the pointer... */
		     "\tsub esp, 30008\n\tmov edi, esp\n\tmov esi, edi\n" \
		     "\tadd esi, 8\n\n\tmov ecx, 30000/4\n\tcld\n" \
		     "\trep stosd\n\n");

	/* write out the program */
	while (pgm->instr != EOP) {
		switch(pgm->instr) {
		case ADDI:
		     	if (pgm->attr == 0) break;
			else if (pgm->attr == 1) fprintf(out, "\tinc al\n");
			else if (pgm->attr == -1) fprintf(out, "\tdec al\n");
			else fprintf(out, "\tadd al, %ld\n", pgm->attr);
			break;
		case PTRI:
		     	if (pgm->attr == 0) break;
			fprintf(out, "\tmov [esi], al\n");
			if (pgm->attr == 1) fprintf(out, "\tinc esi\n");
			else if (pgm->attr == -1) fprintf(out, "\tdec esi\n");
			else fprintf(out, "\tadd esi, %ld\n", pgm->attr);
			fprintf(out, "\tmov al, [esi]\n");
			break;
		case ZERO:
		     	fprintf(out, "\txor eax, eax\n");
			break;
		case IN:
		     	fprintf(out, "\tcall _in\n");
			break;
		case OUT:
		     	fprintf(out, "\tcall _out\n");
			break;
		case ADD:
		     	if (pgm->attr2 == 0) break;
			if (pgm->attr2 == 1)
				fprintf(out, "\tadd [esi+%ld], al\n", \
					pgm->attr);
			else if (pgm->attr2 == -1)
				fprintf(out, "\tsub [esi+%ld], al\n", \
					pgm->attr);
			else if ((pgm->attr == 0) && (pgm->attr2 == -2))
				fprintf(out, "\tneg al\n");
			else fprintf(out, "\timul cx, ax, %ld\n" \
				"\tadd [esi+%ld], cl\n", pgm->attr2, pgm->attr);
			break;
		case LOOP:
		     	/* make sure we have space on the loop stack */
		     	assert(ltop < MAXNEST);
			fprintf(out, "\n\tjmp .l%ldend\n.l%ld\n", \
				label, label);
			lstk[ltop++] = label++;
			break;
		case POOL:
		     	ltop--;
			fprintf(out, ".l%ldend\ttest al, al\n\tjnz .l%ld\n", \
				lstk[ltop], lstk[ltop]);
			break;
		default:
			break;
		}
		pgm++;
	}
	assert(ltop == 0);

		     /* exit the program */
	fprintf(out, "\tmov ebx, eax\n\txor eax, eax\n\tinc eax\n\tint 0x80" \
		     /* input routine */
		     "\n\n_in:\txor ebx, ebx\n\tmov al, 3\n\tpush eax\n" \
		     "\tmov ecx, esp\n\tint 0x80\n\tpop eax\n\tret\n\n" \
		     /* output routine */
		     "_out:\tpush eax\n\tmov ecx, esp\n\tmov al, 4\n" \
		     "\txor ebx, ebx\n\tinc ebx\n\tint 0x80\n\tpop eax\n" \
		     "\tret\n");

	return 0;
}
